"""Export AWS resource tags to XLSX."""

from pathlib import Path

import boto3
import pandas as pd
from rich.console import Console

from .tags import get_tags

console = Console()

TAG_COLUMNS = ["Workload", "Component", "Name", "UUID"]

# Resource type -> (taggable, cost_relevant)
# https://docs.aws.amazon.com/organizations/latest/userguide/orgs_manage_policies_supported-resources-enforcement.html
RESOURCE_TYPES = {
    "acm:certificate": (True, True),
    "apigateway:restapis": (True, True),
    "apigateway:restapis/deployments": (False, False),
    "apigateway:restapis/resources": (False, False),
    "apigateway:restapis/resources/methods": (False, False),
    "apigateway:restapis/stages": (True, False),
    "apprunner:autoscalingconfiguration": (True, False),
    "athena:datacatalog": (True, False),
    "athena:workgroup": (True, True),
    "autoscaling:autoScalingGroup": (
        True,
        True,
    ),  # TODO: use autoscaling:CreateOrUpdateTags
    "backup:backup-plan": (True, False),
    "backup:backup-vault": (True, True),
    "ce:anomalymonitor": (True, False),
    "ce:anomalysubscription": (True, False),
    "cloudformation:stack": (True, True),
    "cloudfront:cache-policy": (False, False),
    "cloudfront:distribution": (True, True),
    "cloudfront:function": (False, False),
    "cloudfront:origin-access-control": (False, False),
    "cloudfront:origin-access-identity": (False, False),
    "cloudtrail:trail": (True, True),
    "cloudwatch:alarm": (True, True),
    "cloudwatch:dashboard": (False, False),
    "codebuild:project": (True, True),
    "codecommit:repository": (True, True),
    "codedeploy:application": (True, False),
    "codepipeline:pipeline": (True, True),
    "datasync:location": (True, False),
    "datasync:task": (True, True),
    "dms:subgrp": (True, False),
    "dynamodb:table": (True, True),
    "ec2:dhcp-options": (True, False),
    "ec2:elastic-ip": (True, True),
    "ec2:fleet": (True, False),
    "ec2:image": (True, True),
    "ec2:instance": (True, True),
    "ec2:internet-gateway": (True, False),
    "ec2:key-pair": (True, False),
    "ec2:launch-template": (True, True),
    "ec2:natgateway": (True, True),
    "ec2:network-acl": (True, False),
    "ec2:network-interface": (True, False),
    "ec2:placement-group": (True, False),
    "ec2:reserved-instances": (True, True),
    "ec2:route-table": (True, False),
    "ec2:security-group": (True, False),
    "ec2:security-group-rule": (True, False),
    "ec2:snapshot": (True, True),
    "ec2:spot-instances-request": (True, True),
    "ec2:subnet": (True, False),
    "ec2:volume": (True, True),
    "ec2:vpc": (True, False),
    "ec2:vpc-endpoint": (True, True),
    "ec2:vpc-peering-connection": (True, False),
    "ecr:repository": (True, True),
    "eks:cluster": (True, True),
    "eks:daemonset": (False, False),
    "eks:deployment": (False, False),
    "eks:endpointslice": (False, False),
    "eks:ingress": (False, False),
    "eks:namespace": (False, False),
    "eks:persistentvolume": (False, False),
    "eks:replicaset": (False, False),
    "eks:service": (False, False),
    "eks:statefulset": (False, False),
    "elasticache:cluster": (True, True),
    "elasticache:parametergroup": (True, True),
    "elasticache:replicationgroup": (True, True),
    "elasticache:reserved-instance": (True, True),
    "elasticache:subnetgroup": (True, True),
    "elasticache:user": (True, True),
    "elasticfilesystem:file-system": (True, True),
    "elasticloadbalancing:listener-rule/app": (True, False),
    "elasticloadbalancing:listener/app": (True, False),
    "elasticloadbalancing:listener/net": (True, False),
    "elasticloadbalancing:loadbalancer": (True, True),
    "elasticloadbalancing:loadbalancer/app": (True, True),
    "elasticloadbalancing:loadbalancer/net": (True, True),
    "elasticloadbalancing:targetgroup": (True, True),
    "es:domain": (True, True),
    "events:event-bus": (True, True),
    "events:rule": (True, True),
    "glue:database": (True, False),
    "glue:table": (False, False),
    "guardduty:detector": (True, True),
    "iam:group": (False, False),
    "iam:instance-profile": (True, False),
    "iam:mfa": (True, False),
    "iam:oidc-provider": (True, False),
    "iam:policy": (True, False),
    "iam:role": (True, False),
    "iam:server-certificate": (True, False),
    "iam:user": (True, False),
    "inspector:target/template": (True, False),
    "kms:key": (True, True),
    "lambda:function": (True, True),
    "logs:log-group": (True, True),
    "memorydb:acl": (True, False),
    "memorydb:parametergroup": (True, False),
    "memorydb:user": (True, False),
    "mq:configuration": (True, False),
    "rds:auto-backup": (False, False),
    "rds:cluster": (True, True),
    "rds:cluster-pg": (True, False),
    "rds:cluster-snapshot": (True, True),
    "rds:db": (True, True),
    "rds:deployment": (True, False),
    "rds:og": (True, False),
    "rds:pg": (True, False),
    "rds:ri": (True, True),
    "rds:secgrp": (True, False),
    "rds:snapshot": (True, True),
    "rds:subgrp": (True, False),
    "resource-explorer-2:index": (False, False),
    "resource-explorer-2:view": (False, False),
    "resource-groups:group": (True, False),
    "route53:healthcheck": (True, True),
    "route53:hostedzone": (True, True),
    "route53resolver:resolver-query-log-config": (True, False),
    "s3:bucket": (True, True),
    "s3:storage-lens": (True, False),
    "secretsmanager:secret": (True, True),
    "ses:configuration-set": (True, False),
    "ses:identity": (True, False),
    "sns:topic": (True, True),
    "sqs:queue": (True, True),
    "ssm:association": (True, False),
    "ssm:document": (True, False),
    "ssm:managed-instance": (True, False),
    "ssm:parameter": (True, True),
    "ssm:resource-data-sync": (False, False),
    "ssm:session": (False, False),
    "wafv2:ipset": (True, False),
    "wafv2:rulegroup": (True, False),
    "wafv2:webacl": (True, True),
}


def get_resources() -> list[dict]:
    """Fetch all resources via Resource Explorer 2."""
    client = boto3.client("resource-explorer-2")
    paginator = client.get_paginator("list_resources")

    resources = []
    for page in paginator.paginate():
        resources.extend(page["Resources"])
        console.print(f"Fetched {len(resources)} resources...")
    return resources


def to_dataframe(resources: list[dict]) -> pd.DataFrame:
    """Convert API response to DataFrame with tag columns."""
    # Filter to taggable and cost-relevant resources
    filtered_resources = []
    for r in resources:
        resource_type = r["ResourceType"]
        if resource_type not in RESOURCE_TYPES:
            raise ValueError(f"Unknown resource type: {resource_type}")
        taggable, cost_relevant = RESOURCE_TYPES[resource_type]
        if taggable and cost_relevant:
            filtered_resources.append(r)

    # Build arn -> resource_type mapping for tag fetcher routing
    arn_to_type = {r["Arn"]: r["ResourceType"] for r in filtered_resources}
    tags = get_tags(arn_to_type) if arn_to_type else {}

    rows = []
    for r in filtered_resources:
        arn = r["Arn"]
        resource_type = r["ResourceType"]
        resource_tags = tags.get(arn, {})

        row = {
            "resource_type": resource_type,
            "resource_arn": arn,
        }
        for tag in TAG_COLUMNS:
            row[f"tag:{tag}"] = resource_tags.get(tag, "")
        rows.append(row)

    return pd.DataFrame(rows).sort_values("resource_arn").reset_index(drop=True)


def to_xlsx(df: pd.DataFrame, output: Path) -> None:
    """Export DataFrame to XLSX."""
    df.to_excel(output, index=False, sheet_name="Resources")



"""CloudWatch metrics fetching with pandas DataFrame output."""

from datetime import datetime
from typing import Literal

import boto3
import pandas as pd
from pydantic import BaseModel, field_validator

from adaptivegears.aws.units import (
    ByteScale,
    ByteRateScale,
    BYTE_FACTORS,
    BYTE_RATE_FACTORS,
)


Stat = Literal["avg", "max", "min", "sum", "count"]

STAT_MAP: dict[Stat, str] = {
    "avg": "Average",
    "max": "Maximum",
    "min": "Minimum",
    "sum": "Sum",
    "count": "SampleCount",
}


def _to_snake_case(name: str) -> str:
    """Convert CamelCase to snake_case, handling acronyms."""
    result = []
    for i, char in enumerate(name):
        if char.isupper() and i > 0:
            # Add underscore before uppercase if previous char is lowercase
            # or if next char is lowercase (end of acronym)
            prev_lower = name[i - 1].islower()
            next_lower = i + 1 < len(name) and name[i + 1].islower()
            if prev_lower or next_lower:
                result.append("_")
        result.append(char.lower())
    return "".join(result)


class Metric(BaseModel):
    """Specification for a CloudWatch metric to fetch."""

    name: str
    stats: list[Stat] = ["avg"]
    scale: ByteScale | ByteRateScale | None = None
    label: str | None = None

    def __init__(
        self,
        name: str | None = None,
        /,
        *,
        stats: list[Stat] | Stat | None = None,
        scale: ByteScale | ByteRateScale | None = None,
        label: str | None = None,
        **kwargs,
    ):
        # Allow positional name argument
        if name is not None:
            kwargs["name"] = name
        if stats is not None:
            kwargs["stats"] = stats
        if scale is not None:
            kwargs["scale"] = scale
        if label is not None:
            kwargs["label"] = label
        super().__init__(**kwargs)

    @field_validator("stats", mode="before")
    @classmethod
    def ensure_list(cls, v):
        if isinstance(v, str):
            return [v]
        return v

    @property
    def column_label(self) -> str:
        """Return the label for DataFrame columns."""
        return self.label or _to_snake_case(self.name)

    def convert(self, value: float) -> float:
        """Apply unit conversion to a value."""
        if self.scale is None:
            return value
        if isinstance(self.scale, ByteScale):
            return value / BYTE_FACTORS[self.scale]
        if isinstance(self.scale, ByteRateScale):
            return value / BYTE_RATE_FACTORS[self.scale]
        return value


def fetch_metrics(
    namespace: str,
    dimensions: dict[str, str],
    metrics: list[Metric],
    start_time: datetime,
    end_time: datetime,
    period: int = 3600,
    region: str = "us-east-1",
) -> pd.DataFrame:
    """
    Fetch CloudWatch metrics and return as a pandas DataFrame.

    Args:
        namespace: CloudWatch namespace (e.g., "AWS/RDS", "AWS/EC2")
        dimensions: Metric dimensions as key-value pairs
        metrics: List of Metric defining what to fetch
        start_time: Start of time range (UTC)
        end_time: End of time range (UTC)
        period: Data point period in seconds (default 3600 = 1 hour)
        region: AWS region

    Returns:
        DataFrame with timestamp index and metric columns named {label}_{stat}
    """
    client = boto3.client("cloudwatch", region_name=region)

    # Build dimensions list
    dims = [{"Name": k, "Value": v} for k, v in dimensions.items()]

    # Build metric queries with counter-based IDs to avoid collisions
    metric_queries = []
    query_map: dict[
        str, tuple[Metric, Stat, str]
    ] = {}  # id -> (metric, stat, col_name)

    for idx, spec in enumerate(metrics):
        for stat_idx, stat in enumerate(spec.stats):
            # Use counter-based ID to guarantee uniqueness
            query_id = f"q_{idx}_{stat_idx}"
            col_name = f"{spec.column_label}_{stat}"

            metric_queries.append(
                {
                    "Id": query_id,
                    "MetricStat": {
                        "Metric": {
                            "Namespace": namespace,
                            "MetricName": spec.name,
                            "Dimensions": dims,
                        },
                        "Period": period,
                        "Stat": STAT_MAP[stat],
                    },
                }
            )
            query_map[query_id] = (spec, stat, col_name)

    # Fetch with pagination
    results: dict[str, list] = {q["Id"]: [] for q in metric_queries}
    timestamps: dict[str, list] = {q["Id"]: [] for q in metric_queries}

    # CloudWatch limits to 500 metrics per request
    chunk_size = 500
    paginator = client.get_paginator("get_metric_data")

    for i in range(0, len(metric_queries), chunk_size):
        chunk = metric_queries[i : i + chunk_size]

        for page in paginator.paginate(
            MetricDataQueries=chunk,
            StartTime=start_time,
            EndTime=end_time,
        ):
            for result in page["MetricDataResults"]:
                query_id = result["Id"]
                results[query_id].extend(result["Values"])
                timestamps[query_id].extend(result["Timestamps"])

    # Build DataFrame using concat to handle sparse data
    series = []
    for query_id, (spec, stat, col_name) in query_map.items():
        ts = timestamps[query_id]
        vals = results[query_id]

        if not ts:
            continue

        converted = [spec.convert(v) for v in vals]
        s = pd.Series(converted, index=ts, name=col_name)
        series.append(s)

    if not series:
        return pd.DataFrame()

    # Concat aligns on index, filling missing values with NaN
    df = pd.concat(series, axis=1)
    df.index.name = "timestamp"
    df = df.sort_index()

    return df



from enum import StrEnum
from typing import Generic, TypeVar

from pydantic import BaseModel

T = TypeVar("T")


class Burstable(BaseModel, Generic[T]):
    """Baseline and max values for burstable resources"""

    baseline: T
    max: T


class ByteScale(StrEnum):
    """Data size scales"""

    # Decimal
    B = "B"
    KB = "KB"
    MB = "MB"
    GB = "GB"
    TB = "TB"
    # Binary
    KiB = "KiB"
    MiB = "MiB"
    GiB = "GiB"
    TiB = "TiB"
    # Bits
    bit = "bit"
    Kbit = "Kbit"
    Mbit = "Mbit"
    Gbit = "Gbit"


class ByteRateScale(StrEnum):
    """Data rate scales"""

    Bps = "Bps"
    KBps = "KBps"
    MBps = "MBps"
    GBps = "GBps"
    bps = "bps"
    Kbps = "Kbps"
    Mbps = "Mbps"
    Gbps = "Gbps"


class CPUScale(StrEnum):
    """CPU scales"""

    core = "core"
    millicore = "millicore"


BYTE_FACTORS: dict[ByteScale, float] = {
    # Decimal
    ByteScale.B: 1,
    ByteScale.KB: 1_000,
    ByteScale.MB: 1_000_000,
    ByteScale.GB: 1_000_000_000,
    ByteScale.TB: 1_000_000_000_000,
    # Binary
    ByteScale.KiB: 2**10,
    ByteScale.MiB: 2**20,
    ByteScale.GiB: 2**30,
    ByteScale.TiB: 2**40,
    # Bits
    ByteScale.bit: 0.125,
    ByteScale.Kbit: 125,
    ByteScale.Mbit: 125_000,
    ByteScale.Gbit: 125_000_000,
}

BYTE_RATE_FACTORS: dict[ByteRateScale, float] = {
    ByteRateScale.Bps: 1,
    ByteRateScale.KBps: 1_000,
    ByteRateScale.MBps: 1_000_000,
    ByteRateScale.GBps: 1_000_000_000,
    ByteRateScale.bps: 0.125,
    ByteRateScale.Kbps: 125,
    ByteRateScale.Mbps: 125_000,
    ByteRateScale.Gbps: 125_000_000,
}

CPU_FACTORS: dict[CPUScale, float] = {
    CPUScale.core: 1,
    CPUScale.millicore: 0.001,
}


class ByteUnit(BaseModel):
    """Data size unit"""

    value: float
    scale: ByteScale

    def to(self, target: ByteScale = ByteScale.B) -> float:
        bytes_ = self.value * BYTE_FACTORS[self.scale]
        return bytes_ / BYTE_FACTORS[target]

    @property
    def bytes(self) -> float:
        return self.to(ByteScale.B)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, ByteUnit):
            return self.bytes < other.bytes
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ByteUnit):
            return self.bytes == other.bytes
        return False

    def __str__(self) -> str:
        return f"{self.value:g} {self.scale}"


class ByteRateUnit(BaseModel):
    """Data rate unit"""

    value: float
    scale: ByteRateScale

    def to(self, target: ByteRateScale = ByteRateScale.Bps) -> float:
        bytes_per_sec = self.value * BYTE_RATE_FACTORS[self.scale]
        return bytes_per_sec / BYTE_RATE_FACTORS[target]

    @property
    def bps(self) -> float:
        return self.to(ByteRateScale.Bps)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, ByteRateUnit):
            return self.bps < other.bps
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ByteRateUnit):
            return self.bps == other.bps
        return False

    def __str__(self) -> str:
        return f"{self.value:g} {self.scale}"


class CPUUnit(BaseModel):
    """CPU unit"""

    value: float
    scale: CPUScale

    def to(self, target: CPUScale = CPUScale.core) -> float:
        cores = self.value * CPU_FACTORS[self.scale]
        return cores / CPU_FACTORS[target]

    @property
    def cores(self) -> float:
        return self.to(CPUScale.core)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, CPUUnit):
            return self.cores < other.cores
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CPUUnit):
            return self.cores == other.cores
        return False

    def __str__(self) -> str:
        return f"{self.value:g} {self.scale}"


class IOPSUnit(BaseModel):
    """IOPS unit (operations per second)"""

    value: int

    def __lt__(self, other: object) -> bool:
        if isinstance(other, IOPSUnit):
            return self.value < other.value
        return NotImplemented

    def __str__(self) -> str:
        return f"{self.value} IOPS"



import boto3
from joblib import Memory
from pydantic import BaseModel

from adaptivegears.aws.region import Region, find_region
from adaptivegears.aws.units import (
    Burstable,
    ByteRateScale,
    ByteRateUnit,
    ByteScale,
    ByteUnit,
    CPUScale,
    CPUUnit,
    IOPSUnit,
)

memory = Memory(".cache", verbose=0)


class Resources(BaseModel):
    """Compute resources specification"""

    cpu: CPUUnit
    memory: ByteUnit
    network: Burstable[ByteRateUnit]
    storage_iops: Burstable[IOPSUnit]
    storage_throughput: Burstable[ByteRateUnit]


class RDSInstance(BaseModel):
    """RDS instance specification"""

    engine: str
    engine_version: str | None
    multi_az: bool
    region: Region
    instance: "EC2Instance"

    @classmethod
    def get(
        cls,
        instance_class: str,
        engine: str,
        engine_version: str | None = None,
        multi_az: bool = False,
        region: str = "us-east-1",
    ) -> "RDSInstance":
        region_model = find_region(region)
        assert region_model, f"Unknown region: {region}"
        instance = EC2Instance.get(instance_class, region=region)
        return cls(
            engine=engine,
            engine_version=engine_version,
            multi_az=multi_az,
            region=region_model,
            instance=instance,
        )


@memory.cache
def fetch_instance_type(instance_type: str, region: str) -> dict:
    """Fetch EC2 instance type info from AWS API (cached)."""
    ec2 = boto3.client("ec2", region_name=region)
    response = ec2.describe_instance_types(InstanceTypes=[instance_type])
    return response["InstanceTypes"][0]


# https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-network-bandwidth.html
class EC2Instance(BaseModel):
    """EC2 instance specification"""

    instance_type: str
    instance_family: str
    architecture: str
    region: str
    resources: Resources

    @classmethod
    def get(cls, instance_type: str, region: str = "us-east-1") -> "EC2Instance":
        """Fetch instance type specs from AWS."""
        # Handle service prefixes (db., cache.) and suffixes (.search)
        ec2_type = (
            instance_type.removeprefix("db.")
            .removeprefix("cache.")
            .removesuffix(".search")
        )
        assert ec2_type.count(".") == 1, f"Invalid instance type: {instance_type}"

        data = fetch_instance_type(ec2_type, region)
        instance_family = ec2_type.split(".")[0]
        architecture = data["ProcessorInfo"]["SupportedArchitectures"][0]

        vcpu = data["VCpuInfo"]["DefaultVCpus"]
        memory_mib = data["MemoryInfo"]["SizeInMiB"]

        network_card = data["NetworkInfo"]["NetworkCards"][0]
        network_baseline = network_card["BaselineBandwidthInGbps"]
        network_max = network_card["PeakBandwidthInGbps"]

        ebs = data["EbsInfo"]["EbsOptimizedInfo"]
        iops_baseline = ebs["BaselineIops"]
        iops_max = ebs["MaximumIops"]
        throughput_baseline = ebs["BaselineThroughputInMBps"]
        throughput_max = ebs["MaximumThroughputInMBps"]

        resources = Resources(
            cpu=CPUUnit(value=vcpu, scale=CPUScale.core),
            memory=ByteUnit(value=memory_mib, scale=ByteScale.MiB),
            network=Burstable(
                baseline=ByteRateUnit(value=network_baseline, scale=ByteRateScale.Gbps),
                max=ByteRateUnit(value=network_max, scale=ByteRateScale.Gbps),
            ),
            storage_iops=Burstable(
                baseline=IOPSUnit(value=iops_baseline),
                max=IOPSUnit(value=iops_max),
            ),
            storage_throughput=Burstable(
                baseline=ByteRateUnit(
                    value=throughput_baseline, scale=ByteRateScale.MBps
                ),
                max=ByteRateUnit(value=throughput_max, scale=ByteRateScale.MBps),
            ),
        )

        return cls(
            instance_type=instance_type,
            instance_family=instance_family,
            architecture=architecture,
            region=region,
            resources=resources,
        )



import json
from typing import Annotated, Any, Literal

import boto3
from pydantic import BaseModel, BeforeValidator, Field, model_validator

from adaptivegears.aws.compute import RDSInstance
from adaptivegears.aws.pricing.offers import Offer, parse_offers


# --- Parsers ---


def parse_int(v: str) -> int:
    """Parse '2' -> 2"""
    return int(v)


# --- Annotated types ---

NormalizationFactor = Annotated[int, BeforeValidator(parse_int)]


# --- Engine mapping ---

ENGINE_MAP = {
    "PostgreSQL": "postgres",
    "MySQL": "mysql",
    "MariaDB": "mariadb",
    "Oracle": "oracle",
    "SQL Server": "sqlserver",
    "Aurora PostgreSQL": "aurora-postgresql",
    "Aurora MySQL": "aurora-mysql",
}


# --- Product Models ---


class RDSInstanceProduct(BaseModel):
    """RDS database instance product with compute specs and pricing."""

    product_family: Literal["Database Instance"]
    sku: str
    deployment_option: str
    normalization_factor: NormalizationFactor = Field(alias="normalizationSizeFactor")
    instance: RDSInstance
    offers: list[Offer]

    @model_validator(mode="before")
    @classmethod
    def flatten_product(cls, data: Any) -> Any:
        assert isinstance(data, dict), f"Expected dict, got {type(data)}"

        product = data.pop("product")
        attrs = product["attributes"]

        data["product_family"] = product["productFamily"]
        data["sku"] = product["sku"]
        data["normalizationSizeFactor"] = attrs["normalizationSizeFactor"]
        data["deployment_option"] = attrs["deploymentOption"]

        instance_type = attrs["instanceType"]
        database_engine = attrs["databaseEngine"]
        deployment_option = attrs["deploymentOption"]
        region_code = attrs["regionCode"]

        engine = ENGINE_MAP.get(database_engine, database_engine.lower())
        multi_az = deployment_option == "Multi-AZ"

        data["instance"] = RDSInstance.get(
            instance_class=instance_type,
            engine=engine,
            multi_az=multi_az,
            region=region_code,
        )

        # Parse offers
        terms = data.pop("terms")
        assert not isinstance(terms, list), "terms already parsed"
        data["offers"] = parse_offers(terms)

        return data


class RDSStorageProduct(BaseModel):
    """RDS database storage product (gp2, gp3, io1, io2).

    Price is per GB (hourly rate converted from GB-Mo).
    Multiply price.daily by storage_gb to get total daily cost.
    """

    product_family: Literal["Database Storage"]
    sku: str
    deployment_option: str
    volume_type: str
    offers: list[Offer]

    @model_validator(mode="before")
    @classmethod
    def flatten_product(cls, data: Any) -> Any:
        assert isinstance(data, dict), f"Expected dict, got {type(data)}"

        product = data.pop("product")
        attrs = product["attributes"]

        data["product_family"] = product["productFamily"]
        data["sku"] = product["sku"]
        data["deployment_option"] = attrs["deploymentOption"]
        data["volume_type"] = attrs["volumeType"]

        terms = data.pop("terms")
        data["offers"] = parse_offers(terms)

        return data


IOPS_GROUP_MAP = {
    "RDS Provisioned IOPS": "io1",
    "RDS Provisioned GP3 IOPS": "gp3",
    "RDS Provisioned IO2 IOPS": "io2",
}


class RDSIOPSProduct(BaseModel):
    """RDS provisioned IOPS product (io1, gp3, io2).

    Price is per IOPS (hourly rate converted from IOPS-Mo).
    Multiply price.daily by iops to get total daily cost.
    """

    product_family: Literal["Provisioned IOPS"]
    sku: str
    deployment_option: str
    volume_type: str
    offers: list[Offer]

    @model_validator(mode="before")
    @classmethod
    def flatten_product(cls, data: Any) -> Any:
        assert isinstance(data, dict), f"Expected dict, got {type(data)}"

        product = data.pop("product")
        attrs = product["attributes"]

        data["product_family"] = product["productFamily"]
        data["sku"] = product["sku"]
        data["deployment_option"] = attrs["deploymentOption"]

        # volumeType only exists for io2, derive from groupDescription
        group_desc = attrs["groupDescription"]
        assert group_desc in IOPS_GROUP_MAP, f"Unknown IOPS group: {group_desc}"
        data["volume_type"] = IOPS_GROUP_MAP[group_desc]

        terms = data.pop("terms")
        data["offers"] = parse_offers(terms)

        return data


class RDSThroughputProduct(BaseModel):
    """RDS provisioned throughput product (gp3).

    Price is per MiBps (hourly rate converted from MBPS-Mo).
    Multiply price.daily by throughput_mibps to get total daily cost.
    """

    product_family: Literal["Provisioned Throughput"]
    sku: str
    deployment_option: str
    offers: list[Offer]

    @model_validator(mode="before")
    @classmethod
    def flatten_product(cls, data: Any) -> Any:
        assert isinstance(data, dict), f"Expected dict, got {type(data)}"

        product = data.pop("product")
        attrs = product["attributes"]

        data["product_family"] = product["productFamily"]
        data["sku"] = product["sku"]
        data["deployment_option"] = attrs["deploymentOption"]

        terms = data.pop("terms")
        data["offers"] = parse_offers(terms)

        return data


# --- Discriminated Union ---

RDSProduct = Annotated[
    RDSInstanceProduct | RDSStorageProduct | RDSIOPSProduct | RDSThroughputProduct,
    Field(discriminator="product_family"),
]

PRODUCT_CLASS_MAP = {
    "Database Instance": RDSInstanceProduct,
    "Database Storage": RDSStorageProduct,
    "Provisioned IOPS": RDSIOPSProduct,
    "Provisioned Throughput": RDSThroughputProduct,
}


# --- Pricing Client ---


class Pricing:
    def __init__(self, region_name="us-east-1"):
        self.client = boto3.client("pricing", region_name=region_name)

    def get_rds_products(self, filters) -> list[RDSProduct]:
        """Fetch RDS products matching filters with pagination."""
        paginator = self.client.get_paginator("get_products")
        products = []

        for page in paginator.paginate(ServiceCode="AmazonRDS", Filters=filters):
            for item in page["PriceList"]:
                raw = json.loads(item)
                product_family = raw["product"]["productFamily"]
                product_class = PRODUCT_CLASS_MAP.get(product_family)
                if product_class:
                    products.append(product_class.model_validate(raw))

        return products



import typer

from adaptivegears import __version__
from adaptivegears.aws.cli import app as aws_app
from adaptivegears.lookup.cli import app as lookup_app
from adaptivegears.pg.cli import app as pg_app
from adaptivegears.tags.cli import app as tags_app
from adaptivegears.util.cli import _uuid7

app = typer.Typer(help="AdaptiveGears CLI tools")


def version_callback(value: bool):
    if value:
        print(f"adaptivegears {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        False, "--version", "-v", callback=version_callback, is_eager=True
    ),
):
    pass


@app.command()
def uuid(
    v7: bool = typer.Option(False, "-7", help="Generate UUID v7 (time-sortable)"),
    count: int = typer.Option(1, "-n", help="Number of UUIDs to generate"),
):
    """Generate UUIDs."""
    import uuid as uuid_lib

    for _ in range(count):
        if v7:
            print(_uuid7())
        else:
            print(uuid_lib.uuid4())


# Register subcommands
app.add_typer(aws_app, name="aws")
app.add_typer(lookup_app, name="lookup")
app.add_typer(pg_app, name="pg")
app.add_typer(tags_app, name="tags")



import boto3
from joblib import Memory
from pydantic import BaseModel

memory = Memory(".cache", verbose=0)

# https://docs.aws.amazon.com/global-infrastructure/latest/regions/aws-region-billing-codes.html
BILLING_CODES = {
    "us-east-1": "USE1",
    "us-east-2": "USE2",
    "us-west-1": "USW1",
    "us-west-2": "USW2",
    "ca-central-1": "CAN1",
    "ca-west-1": "CAN2",
    "mx-central-1": "MXC1",
    "af-south-1": "AFS1",
    "ap-east-1": "APE1",
    "ap-east-2": "APE2",
    "ap-northeast-1": "APN1",
    "ap-northeast-2": "APN2",
    "ap-northeast-3": "APN3",
    "ap-south-1": "APS3",
    "ap-south-2": "APS5",
    "ap-southeast-1": "APS1",
    "ap-southeast-2": "APS2",
    "ap-southeast-3": "APS4",
    "ap-southeast-4": "APS6",
    "ap-southeast-5": "APS7",
    "ap-southeast-6": "APS8",
    "ap-southeast-7": "APS9",
    "eu-central-1": "EUC1",
    "eu-central-2": "EUC2",
    "eu-north-1": "EUN1",
    "eu-south-1": "EUS1",
    "eu-south-2": "EUS2",
    "eu-west-1": "EU",
    "eu-west-2": "EUW2",
    "eu-west-3": "EUW3",
    "il-central-1": "ILC1",
    "me-central-1": "MEC1",
    "me-south-1": "MES1",
    "sa-east-1": "SAE1",
    # GovCloud
    "us-gov-west-1": "UGW1",
    "us-gov-east-1": "UGE1",
    # China
    "cn-north-1": "CNN1",
    "cn-northwest-1": "CNW1",
    # European Sovereign Cloud (new partition, billing code TBD)
    "eusc-de-east-1": "EUSCDE1",
}


class Region(BaseModel):
    code: str
    name: str
    abbrv: str

    @property
    def pricing_filter(self) -> dict:
        return {"Type": "TERM_MATCH", "Field": "regionCode", "Value": self.code}


@memory.cache
def fetch_regions() -> list[dict]:
    """Fetch all AWS regions from SSM global infrastructure."""
    ssm = boto3.client("ssm", region_name="us-east-1")

    regions = []
    paginator = ssm.get_paginator("get_parameters_by_path")

    for page in paginator.paginate(Path="/aws/service/global-infrastructure/regions"):
        for param in page["Parameters"]:
            code = param["Value"]

            r = ssm.get_parameter(
                Name=f"/aws/service/global-infrastructure/regions/{code}/longName"
            )
            name = r["Parameter"]["Value"]
            assert code in BILLING_CODES, f"Missing billing code for region: {code}"
            abbrv = BILLING_CODES[code].lower()

            regions.append({"code": code, "name": name, "abbrv": abbrv})

    return regions


def get_regions() -> list[Region]:
    """Get cached regions as Pydantic models."""
    regions = [Region(**r) for r in fetch_regions()]
    assert regions, "No regions fetched from AWS"
    return regions


def find_region(query: str) -> Region | None:
    """Find region by code, name, or abbreviation (case-insensitive)."""
    query = query.lower()
    for r in get_regions():
        if r.code.lower() == query:
            return r
        if r.name.lower() == query:
            return r
        if r.abbrv == query:
            return r
    return None



"""CLI commands for resource lookup."""

import json

import boto3
import typer

from .ami import build_report as build_ami_report
from .ami import delete_ami
from .lc import build_report as build_lc_report
from .lt import build_report as build_lt_report

app = typer.Typer(help="Resource lookup tools")


@app.command("lt")
def lt_report(
    orphaned: bool = typer.Option(
        False, "--orphaned", help="Show only orphaned launch templates"
    ),
    sort: str = typer.Option("name", "-s", "--sort", help="Sort by: name, created"),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON"),
    region: str = typer.Option(None, "-r", "--region", help="AWS region"),
):
    """Launch Template usage report."""
    session = boto3.Session(region_name=region)
    report = build_lt_report(session)

    # Sort
    if sort == "name":
        report.sort(key=lambda r: r.name)
    elif sort == "created":
        report.sort(key=lambda r: r.created)
    else:
        typer.echo(f"Unknown sort column: {sort}. Use 'name' or 'created'.")
        raise typer.Exit(1)

    if orphaned:
        report = [r for r in report if r.is_orphaned]

    if json_output:
        print(json.dumps([r.to_dict() for r in report], indent=2))
    else:
        if not report:
            typer.echo("No launch templates found.")
            return

        # Calculate column widths
        name_width = max(len(r.name) for r in report)
        name_width = max(name_width, len("Launch Template"))

        # Header
        typer.echo(
            f"{'Launch Template':<{name_width}}  {'Created':<12}  {'EC2':>5}  References"
        )
        typer.echo(f"{'-' * name_width}  {'-' * 12}  {'-' * 5}  {'-' * 40}")

        # Rows
        for r in report:
            created_str = r.created.strftime("%Y-%m-%d")
            if r.references:
                refs_str = ", ".join(f"{ref.type}:{ref.name}" for ref in r.references)
            else:
                refs_str = "-"
            typer.echo(
                f"{r.name:<{name_width}}  {created_str:<12}  {r.instance_count:>5}  {refs_str}"
            )

        # Summary
        orphaned_count = sum(1 for r in report if r.is_orphaned)
        typer.echo()
        typer.echo(f"Total: {len(report)} launch templates, {orphaned_count} orphaned")


@app.command("lc")
def lc_report(
    orphaned: bool = typer.Option(
        False, "--orphaned", help="Show only orphaned launch configurations"
    ),
    sort: str = typer.Option("name", "-s", "--sort", help="Sort by: name, created"),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON"),
    region: str = typer.Option(None, "-r", "--region", help="AWS region"),
):
    """Launch Configuration usage report."""
    session = boto3.Session(region_name=region)
    report = build_lc_report(session)

    # Sort
    if sort == "name":
        report.sort(key=lambda r: r.name)
    elif sort == "created":
        report.sort(key=lambda r: r.created)
    else:
        typer.echo(f"Unknown sort column: {sort}. Use 'name' or 'created'.")
        raise typer.Exit(1)

    if orphaned:
        report = [r for r in report if r.is_orphaned]

    if json_output:
        print(json.dumps([r.to_dict() for r in report], indent=2))
    else:
        if not report:
            typer.echo("No launch configurations found.")
            return

        # Calculate column widths
        name_width = max(len(r.name) for r in report)
        name_width = max(name_width, len("Launch Configuration"))

        # Header
        typer.echo(
            f"{'Launch Configuration':<{name_width}}  {'Created':<12}  {'AMI ID':<21}  References"
        )
        typer.echo(f"{'-' * name_width}  {'-' * 12}  {'-' * 21}  {'-' * 40}")

        # Rows
        for r in report:
            created_str = r.created.strftime("%Y-%m-%d")
            refs_str = (
                ", ".join(f"{ref.type}:{ref.name}" for ref in r.references)
                if r.references
                else "-"
            )
            typer.echo(
                f"{r.name:<{name_width}}  {created_str:<12}  {r.ami_id:<21}  {refs_str}"
            )

        # Summary
        orphaned_count = sum(1 for r in report if r.is_orphaned)
        typer.echo()
        typer.echo(
            f"Total: {len(report)} launch configurations, {orphaned_count} orphaned"
        )


@app.command("ami")
def ami_report(
    orphaned: bool = typer.Option(False, "--orphaned", help="Show only orphaned AMIs"),
    delete: bool = typer.Option(
        False, "--delete", help="Interactive deletion (requires --orphaned)"
    ),
    sort: str = typer.Option("name", "-s", "--sort", help="Sort by: name, created"),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON"),
    region: str = typer.Option(None, "-r", "--region", help="AWS region"),
):
    """AMI usage report."""
    if delete and not orphaned:
        typer.echo("--delete requires --orphaned")
        raise typer.Exit(1)

    session = boto3.Session(region_name=region)
    report = build_ami_report(session)

    # Sort
    if sort == "name":
        report.sort(key=lambda r: r.name)
    elif sort == "created":
        report.sort(key=lambda r: r.created)
    else:
        typer.echo(f"Unknown sort column: {sort}. Use 'name' or 'created'.")
        raise typer.Exit(1)

    if orphaned:
        report = [r for r in report if r.is_orphaned]

    if json_output:
        print(json.dumps([r.to_dict() for r in report], indent=2))
    else:
        if not report:
            typer.echo("No AMIs found.")
            return

        # Calculate column widths
        name_width = max(len(r.name) for r in report)
        name_width = max(name_width, len("AMI Name"))

        # Header
        typer.echo(
            f"{'AMI Name':<{name_width}}  {'AMI ID':<21}  {'Created':<12}  {'EC2':>5}  References  Snapshots"
        )
        typer.echo(f"{'-' * name_width}  {'-' * 21}  {'-' * 12}  {'-' * 5}  {'-' * 40}")

        # Rows
        for r in report:
            created_str = r.created.strftime("%Y-%m-%d")
            refs_str = (
                ", ".join(f"{ref.type}:{ref.name}" for ref in r.references)
                if r.references
                else "-"
            )
            snaps_str = ", ".join(r.snapshot_ids) if r.snapshot_ids else "-"
            typer.echo(
                f"{r.name:<{name_width}}  {r.id:<21}  {created_str:<12}  {r.instance_count:>5}  {refs_str}  {snaps_str}"
            )

        # Summary
        orphaned_count = sum(1 for r in report if r.is_orphaned)
        total_snapshots = sum(len(r.snapshot_ids) for r in report)
        typer.echo()
        typer.echo(
            f"Total: {len(report)} AMIs ({total_snapshots} snapshots), {orphaned_count} orphaned"
        )

        # Interactive deletion
        if delete and report:
            typer.echo()
            typer.echo("Starting interactive deletion (y=delete, n=skip, q=quit)")
            typer.echo()

            deleted_amis = 0
            deleted_snaps = 0

            for r in report:
                created_str = r.created.strftime("%Y-%m-%d")
                total_size = sum(s.size_gb for s in r.snapshots)
                typer.echo(f"AMI: {r.name}")
                typer.echo(f"  ID:      {r.id}")
                typer.echo(f"  Created: {created_str}")
                if r.snapshots:
                    typer.echo(f"  Snapshots ({len(r.snapshots)}, {total_size} GB):")
                    for snap in r.snapshots:
                        typer.echo(f"    - {snap.id} ({snap.size_gb} GB)")
                else:
                    typer.echo("  Snapshots: none")

                while True:
                    choice = typer.prompt("Delete?", default="y")
                    if choice.lower() == "y":
                        delete_ami(session, r.id, r.snapshot_ids)
                        deleted_amis += 1
                        deleted_snaps += len(r.snapshot_ids)
                        break
                    elif choice.lower() == "n":
                        typer.echo("  Skipped")
                        break
                    elif choice.lower() == "q":
                        typer.echo()
                        typer.echo(
                            f"Deleted {deleted_amis} AMIs, {deleted_snaps} snapshots"
                        )
                        return
                    else:
                        typer.echo("  Invalid choice. Use y/n/q")

                typer.echo()

            typer.echo(f"Deleted {deleted_amis} AMIs, {deleted_snaps} snapshots")



import typer

from adaptivegears.aws.compute import EC2Instance

app = typer.Typer(help="AWS tools")
ec2_app = typer.Typer(help="EC2 instance tools")
app.add_typer(ec2_app, name="ec2")


@ec2_app.command("get")
def ec2_get(
    instance_type: str = typer.Argument(..., help="Instance type (e.g., r6g.large)"),
    region: str = typer.Option("us-east-1", "-r", "--region", help="AWS region"),
    json: bool = typer.Option(False, "--json", help="Output as JSON"),
):
    """Get EC2 instance specifications."""
    instance = EC2Instance.get(instance_type, region=region)

    if json:
        print(instance.model_dump_json(indent=2))
    else:
        r = instance.resources
        print(f"Instance: {instance.instance_type} ({instance.region})")
        print(f"  Family:     {instance.instance_family}")
        print(f"  Arch:       {instance.architecture}")
        print(f"  CPU:        {r.cpu.value:.0f} {r.cpu.scale}")
        print(f"  Memory:     {r.memory.to(r.memory.scale):.0f} {r.memory.scale}")
        print(
            f"  Network:    {r.network.baseline.value:.2f}-{r.network.max.value:.2f} {r.network.baseline.scale}"
        )
        print(
            f"  EBS IOPS:   {r.storage_iops.baseline.value}-{r.storage_iops.max.value}"
        )
        print(
            f"  EBS Tput:   {r.storage_throughput.baseline.value:.2f}-{r.storage_throughput.max.value:.2f} {r.storage_throughput.baseline.scale}"
        )



"""Tag fetching and setting with service-specific API support."""

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Callable

import boto3
from rich.progress import track


@dataclass
class TagChange:
    """Represents a tag change to apply to a resource."""

    arn: str
    resource_type: str
    tags_to_set: dict[str, str]


@dataclass
class SetTagsResult:
    """Result of applying tag changes."""

    success: list[str] = field(default_factory=list)
    failed: dict[str, str] = field(default_factory=dict)  # ARN -> error message


def get_tags_standard(arns: list[str], batch_size: int = 100) -> dict[str, dict]:
    """Fetch tags via ResourceGroupsTaggingAPI (batch of 100)."""
    if not arns:
        return {}

    client = boto3.client("resourcegroupstaggingapi")
    tags = {}
    batches = list(range(0, len(arns), batch_size))

    for start in track(batches, description="Fetching tags (standard)..."):
        batch = arns[start : start + batch_size]
        response = client.get_resources(ResourceARNList=batch)
        for r in response["ResourceTagMappingList"]:
            tags[r["ResourceARN"]] = {t["Key"]: t["Value"] for t in r.get("Tags", [])}

    return tags


def get_tags_iam(arns: list[str]) -> dict[str, dict]:
    """Fetch tags for IAM resources via dedicated API.

    - iam:ListRoleTags for roles
    - iam:ListUserTags for users
    """
    if not arns:
        return {}

    client = boto3.client("iam")
    tags = {}

    for arn in track(arns, description="Fetching tags (IAM)..."):
        # arn:aws:iam::123456789012:role/MyRole
        # arn:aws:iam::123456789012:user/MyUser
        parts = arn.split(":")
        if len(parts) < 6:
            continue

        resource = parts[5]  # role/MyRole or role/path/MyRole
        if "/" not in resource:
            continue

        # IAM resources can have paths: role/service-role/MyRole
        # API wants just the name (last segment), not the path
        resource_parts = resource.split("/")
        resource_type = resource_parts[0]  # "role" or "user"
        resource_name = resource_parts[-1]  # actual name (last segment)

        try:
            if resource_type == "role":
                response = client.list_role_tags(RoleName=resource_name)
                tags[arn] = {t["Key"]: t["Value"] for t in response.get("Tags", [])}
            elif resource_type == "user":
                response = client.list_user_tags(UserName=resource_name)
                tags[arn] = {t["Key"]: t["Value"] for t in response.get("Tags", [])}
        except client.exceptions.NoSuchEntityException:
            # Resource was deleted between listing and tag fetch
            pass

    return tags


def get_tags_wafv2(arns: list[str]) -> dict[str, dict]:
    """Fetch tags for WAFv2 resources via dedicated API.

    - wafv2:ListTagsForResource
    """
    if not arns:
        return {}

    client = boto3.client("wafv2")
    tags = {}

    for arn in track(arns, description="Fetching tags (WAFv2)..."):
        try:
            response = client.list_tags_for_resource(ResourceARN=arn)
            tag_list = response.get("TagInfoForResource", {}).get("TagList", [])
            tags[arn] = {t["Key"]: t["Value"] for t in tag_list}
        except client.exceptions.WAFNonexistentItemException:
            # Resource was deleted between listing and tag fetch
            pass

    return tags


def get_tags_autoscaling(arns: list[str]) -> dict[str, dict]:
    """Fetch tags for Auto Scaling Groups via dedicated API.

    - autoscaling:DescribeTags
    """
    if not arns:
        return {}

    client = boto3.client("autoscaling")
    tags = {}

    # Extract ASG names from ARNs
    # arn:aws:autoscaling:region:account:autoScalingGroup:uuid:autoScalingGroupName/name
    asg_names = []
    arn_to_name = {}
    for arn in arns:
        parts = arn.split(":")
        if len(parts) >= 8:
            # Last part is autoScalingGroupName/actual-name
            name_part = parts[7]
            if name_part.startswith("autoScalingGroupName/"):
                name = name_part.split("/", 1)[1]
                asg_names.append(name)
                arn_to_name[arn] = name

    if not asg_names:
        return {}

    # Fetch tags for all ASGs (API returns tags for all matching filters)
    paginator = client.get_paginator("describe_tags")
    name_to_tags: dict[str, dict] = defaultdict(dict)

    for page in paginator.paginate(
        Filters=[{"Name": "auto-scaling-group", "Values": asg_names}]
    ):
        for tag in page["Tags"]:
            asg_name = tag["ResourceId"]
            name_to_tags[asg_name][tag["Key"]] = tag["Value"]

    # Map back to ARNs
    for arn, name in arn_to_name.items():
        tags[arn] = name_to_tags.get(name, {})

    return tags


# =============================================================================
# Tag Setters
# =============================================================================


def set_tags_standard(arn_to_tags: dict[str, dict[str, str]]) -> dict[str, bool]:
    """Set tags via ResourceGroupsTaggingAPI.

    Note: tag_resources applies same tags to all ARNs in batch,
    so we call per-resource for flexibility.
    """
    if not arn_to_tags:
        return {}

    client = boto3.client("resourcegroupstaggingapi")
    results = {}

    for arn, tags_dict in track(
        arn_to_tags.items(), description="Setting tags (standard)..."
    ):
        try:
            client.tag_resources(
                ResourceARNList=[arn],
                Tags=tags_dict,
            )
            results[arn] = True
        except Exception:
            results[arn] = False

    return results


def set_tags_iam(arn_to_tags: dict[str, dict[str, str]]) -> dict[str, bool]:
    """Set tags for IAM resources via dedicated API.

    - iam:TagRole
    - iam:TagUser
    """
    if not arn_to_tags:
        return {}

    client = boto3.client("iam")
    results = {}

    for arn, tags_dict in track(
        arn_to_tags.items(), description="Setting tags (IAM)..."
    ):
        parts = arn.split(":")
        if len(parts) < 6:
            results[arn] = False
            continue

        resource = parts[5]
        if "/" not in resource:
            results[arn] = False
            continue

        resource_parts = resource.split("/")
        resource_type = resource_parts[0]
        resource_name = resource_parts[-1]

        tags_list = [{"Key": k, "Value": v} for k, v in tags_dict.items()]

        try:
            if resource_type == "role":
                client.tag_role(RoleName=resource_name, Tags=tags_list)
                results[arn] = True
            elif resource_type == "user":
                client.tag_user(UserName=resource_name, Tags=tags_list)
                results[arn] = True
            else:
                results[arn] = False
        except Exception:
            results[arn] = False

    return results


def set_tags_wafv2(arn_to_tags: dict[str, dict[str, str]]) -> dict[str, bool]:
    """Set tags for WAFv2 resources via wafv2:TagResource."""
    if not arn_to_tags:
        return {}

    client = boto3.client("wafv2")
    results = {}

    for arn, tags_dict in track(
        arn_to_tags.items(), description="Setting tags (WAFv2)..."
    ):
        tags_list = [{"Key": k, "Value": v} for k, v in tags_dict.items()]

        try:
            client.tag_resource(ResourceARN=arn, Tags=tags_list)
            results[arn] = True
        except Exception:
            results[arn] = False

    return results


def set_tags_autoscaling(arn_to_tags: dict[str, dict[str, str]]) -> dict[str, bool]:
    """Set tags for Auto Scaling Groups via autoscaling:CreateOrUpdateTags."""
    if not arn_to_tags:
        return {}

    client = boto3.client("autoscaling")
    results = {}

    for arn, tags_dict in track(
        arn_to_tags.items(), description="Setting tags (ASG)..."
    ):
        # Extract ASG name from ARN
        parts = arn.split(":")
        if len(parts) < 8:
            results[arn] = False
            continue

        name_part = parts[7]
        if not name_part.startswith("autoScalingGroupName/"):
            results[arn] = False
            continue

        asg_name = name_part.split("/", 1)[1]

        tags_list = [
            {
                "ResourceId": asg_name,
                "ResourceType": "auto-scaling-group",
                "Key": k,
                "Value": v,
                "PropagateAtLaunch": True,
            }
            for k, v in tags_dict.items()
        ]

        try:
            client.create_or_update_tags(Tags=tags_list)
            results[arn] = True
        except Exception:
            results[arn] = False

    return results


# Fetcher type alias
Fetcher = Callable[[list[str]], dict[str, dict]]

# Map resource_type -> fetcher (default is standard via ResourceGroupsTaggingAPI)
RESOURCE_TAG_GETTERS: dict[str, Fetcher] = {
    "iam:role": get_tags_iam,
    "iam:user": get_tags_iam,
    "wafv2:ipset": get_tags_wafv2,
    "wafv2:rulegroup": get_tags_wafv2,
    "wafv2:webacl": get_tags_wafv2,
    "autoscaling:autoScalingGroup": get_tags_autoscaling,
}

# Setter type alias
Setter = Callable[[dict[str, dict[str, str]]], dict[str, bool]]

# Map resource_type -> setter (default is standard via ResourceGroupsTaggingAPI)
RESOURCE_TAG_SETTERS: dict[str, Setter] = {
    "iam:role": set_tags_iam,
    "iam:user": set_tags_iam,
    "wafv2:ipset": set_tags_wafv2,
    "wafv2:rulegroup": set_tags_wafv2,
    "wafv2:webacl": set_tags_wafv2,
    "autoscaling:autoScalingGroup": set_tags_autoscaling,
}


def get_tags(arn_to_type: dict[str, str], batch_size: int = 100) -> dict[str, dict]:
    """Fetch tags for ARNs, routing by resource type to appropriate fetcher.

    Args:
        arn_to_type: Mapping of ARN -> resource_type
        batch_size: Batch size for standard API (max 100)

    Returns:
        Mapping of ARN -> tags dict
    """
    if not arn_to_type:
        raise ValueError("arn_to_type cannot be empty")

    # Group ARNs by fetcher
    standard_arns: list[str] = []
    custom_groups: dict[Fetcher, list[str]] = defaultdict(list)

    for arn, resource_type in arn_to_type.items():
        fetcher = RESOURCE_TAG_GETTERS.get(resource_type)
        if fetcher:
            custom_groups[fetcher].append(arn)
        else:
            standard_arns.append(arn)

    # Fetch via standard API
    tags = get_tags_standard(standard_arns, batch_size)

    # Fetch via custom fetchers
    for fetcher, arns in custom_groups.items():
        tags.update(fetcher(arns))

    return tags


def set_tags(changes: list[TagChange]) -> SetTagsResult:
    """Apply tag changes, routing by resource type to appropriate setter.

    Args:
        changes: List of TagChange objects

    Returns:
        SetTagsResult with success/failed ARNs
    """
    if not changes:
        return SetTagsResult()

    # Group changes by setter
    standard_changes: dict[str, dict[str, str]] = {}
    custom_groups: dict[Setter, dict[str, dict[str, str]]] = defaultdict(dict)

    for change in changes:
        setter = RESOURCE_TAG_SETTERS.get(change.resource_type)
        if setter:
            custom_groups[setter][change.arn] = change.tags_to_set
        else:
            standard_changes[change.arn] = change.tags_to_set

    result = SetTagsResult()

    # Apply via standard API
    if standard_changes:
        standard_results = set_tags_standard(standard_changes)
        for arn, success in standard_results.items():
            if success:
                result.success.append(arn)
            else:
                result.failed[arn] = "Failed to set tags"

    # Apply via custom setters
    for setter, arn_to_tags in custom_groups.items():
        setter_results = setter(arn_to_tags)
        for arn, success in setter_results.items():
            if success:
                result.success.append(arn)
            else:
                result.failed[arn] = "Failed to set tags"

    return result
